#include <mysql.h>
#include <uuid/uuid.h>
#include <netinet/in.h>
#include <openssl/evp.h>
#include <syslog.h>

#include "config.h"
#include "rzb_global.h"
#include "rzb_conf.h"
#include "rzb_dispatcher.h"
#include "rzb_component_global.h"
#include "rzb_component_db.h"
#include "rzb_alert_global.h"
#include "rzb_alert_db.h"
#include "rzb_db.h"
#include "rzb_client.h"
#include "rzb_management_log.h"

// SQL Definitions for General Alerting
static const char GET_MESSAGE_BY_ID[] = "SELECT Message from alerts where Alert_ID = %u";
static const char GET_SRC_DATA_BY_ID[] = "SELECT Data from components where Component_ID = %u";
static const char GET_COMP_ID_BY_ALERT_ID[] = "SELECT %s from alerts where Alert_ID = %u";

// SQL Definitions for Inserting Blocks of Data
static const char UPDATE_WITH_DATA[] = "UPDATE components SET Data = unhex('%s') WHERE Component_ID = %u";

static const char INSERT_ALERT_SQL[] = "INSERT INTO alerts (Event_ID, Nugget_ID, Timestamp, Priority, Message, Src_IP, Dst_IP, IP_Proto, Src_Port, Dst_port) VALUES (?, ?, FROM_UNIXTIME(?), ?, ?, ?, ?, ?, ?, ?)";
#define INSERT_ALERT_SQL_ARGC    10

static const char INSERT_BLOB_SQL[] = "UPDATE alerts SET %s = ? WHERE Alert_ID = ?";
#define INSERT_BLOB_SQL_ARGC 2

static const char INSERT_NEW_COMPONENT_SQL[] = "INSERT INTO components (Checksum, Data, Size, Datatype) VALUES ('%s', unhex('%s'), %u, '%.*s')";
static const char INSERT_NEW_COMPONENT_BLOCK_SQL[] = "INSERT INTO component_blocks (Main_Component) VALUES ('%u')";
static const char UPDATE_COMPONENT_WITH_BLOCK_SQL[] = "UPDATE components SET Block_ID=%u WHERE Component_ID =%u";
static const char COMPONENT_CHECK_SQL[] = "SELECT Block_ID, Component_ID FROM components WHERE (Checksum = '%s' AND Size = %u)";
static const char UPDATE_ALERT_WITH_DATA_SQL[] =  "UPDATE alerts SET %s=%u, Full_Block=%u WHERE Alert_ID = %u";
static const char UPDATE_COMPONENT_WITH_NORM_SQL[] =  "UPDATE components SET Norm_ID=%u WHERE Component_ID=%u";
static const char RENDER_VERDICT[] = "UPDATE components SET Threat_flags = (Threat_flags | %u) WHERE Component_ID=%u";

static HRESULT getComponentIDByAlertID(const char *type, unsigned id, unsigned *c_id)
{
    MYSQL *conn;
    MYSQL_RES *response;
    MYSQL_ROW row;
    unsigned row_count;

    char *request;
    size_t request_size;

    *c_id = 0;
    request_size = sizeof(GET_COMP_ID_BY_ALERT_ID) + strlen(type) + 2 + MAX_INT_STRING_SIZE;
    if ((request = malloc(request_size)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate memory for request", "getComponentIDByAlertID");
        return(R_FAIL);
    }

    snprintf(request, request_size, GET_COMP_ID_BY_ALERT_ID, type, id);

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        return R_FAIL;
    }
    if (mysql_query(conn, request) !=0)
    {
        printf("Query failed: %s\n", request);
        printError("Failed results check in componentCheck\n", conn);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(request);

    // Check response.  If there are no rows, it isn't in the databse
    response = mysql_store_result(conn);
    row_count = mysql_num_rows(response);

    if (row_count == 0)
    {
        printf("Returning component not in db\n");
        mysql_free_result(response);
        putMYSQLConnection(conn);
        return R_NOT_FOUND;
    }

    row = mysql_fetch_row(response);
    *c_id = strtoul(row[0], NULL, 10);
    mysql_free_result(response);
    putMYSQLConnection(conn);
    return R_SUCCESS;
}

HRESULT getDataByAlertID(unsigned type, unsigned char **data, unsigned *size, int id)
{
    static const char GET_ALERT_DATA_BY_ID[] = "SELECT %s from alerts where Alert_ID = %u";
    MYSQL *conn;
    MYSQL_RES *response;
    MYSQL_ROW row;

    char *request;
    const char *str_var = NULL;
    size_t request_size;
    unsigned row_count;
    unsigned component_id;
    unsigned long *lengths;

    switch (type)
    {
        case SHORT_DATA:
            str_var = SHORT_DATA_STRING;
        case LONG_DATA:
            if (!str_var)
                str_var = LONG_DATA_STRING;
            request_size = strlen(str_var) + sizeof(GET_ALERT_DATA_BY_ID) + MAX_INT_STRING_SIZE;
            request = malloc(request_size);
            if (!request)
            {
                submitLog(LOG_ERR, "Unable to allocate space for request", "getDataByAlertID");
                return R_FAIL;
            }
            snprintf(request, request_size, GET_ALERT_DATA_BY_ID, str_var, id);
            break;
        case FULL_BLOCK:
            str_var = FULL_ALERT_FIELD;
        case NORM_BLOCK:
            if (!str_var)
                str_var = NORM_ALERT_FIELD;
            getComponentIDByAlertID(str_var, id, &component_id);
            request_size = sizeof(GET_SRC_DATA_BY_ID) + MAX_INT_STRING_SIZE;
            request = malloc(request_size);
            if (!request)
            {
                submitLog(LOG_ERR, "Unable to allocate space for request", "getDataByAlertID");
                return R_FAIL;
            }
            snprintf(request, request_size, GET_SRC_DATA_BY_ID, component_id);
            break;
        case ALRT_MESSAGE:
            request_size = sizeof(GET_MESSAGE_BY_ID) + MAX_INT_STRING_SIZE;
            request = malloc(request_size);
            if (!request)
            {
                submitLog(LOG_ERR, "Unable to allocate space for request", "getDataByAlertID");
                return R_FAIL;
            }
            snprintf(request, request_size, GET_MESSAGE_BY_ID, id);
            break;
        default:
            submitLog(LOG_ERR, "Invalid blob type", "getDataByAlertID");
            return R_FAIL;
    }

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        return R_FAIL;
    }
    if (mysql_query(conn, request) !=0)
    {
        printf("Query failed: %s\n", request);
        printError("Failed results check in componentCheck\n", conn);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(request);

    // Check response.  If there are no rows, it isn't in the databse
    response = mysql_store_result(conn);
    row_count = mysql_num_rows(response);

    if (!row_count)
    {
        printf("Returning component not in db\n");
        mysql_free_result (response);
        putMYSQLConnection(conn);
        return R_NOT_FOUND;
    }

    row = mysql_fetch_row(response);
    lengths = mysql_fetch_lengths(response);

    *size = lengths[0];
    if (*size == 0)
    {
        mysql_free_result(response);
        putMYSQLConnection(conn);
        return R_SUCCESS;
    }

    if ((*data = malloc(*size)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate space for data", "getDataByAlertID");
        mysql_free_result(response);
        putMYSQLConnection(conn);
        return R_MALLOC_FAIL;
    }

    memcpy(*data, row[0], *size);
    mysql_free_result(response);
    putMYSQLConnection(conn);
    return R_SUCCESS;
}

HRESULT renderVerdict(JUDGEMENT *verdict)
{
    MYSQL *conn;
    size_t request_size;
    char *request;
    COMPONENT *component;

    if ((component = malloc(sizeof(COMPONENT))) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate memory for component", "renderVerdict");
        return R_MALLOC_FAIL;
    }

    if (getComponent(verdict->md5, verdict->size, component) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to get component id", "renderVerdict");
        free(component);
        return R_FAIL;
    }

    request_size = sizeof(RENDER_VERDICT) + (2 * MAX_INT_STRING_SIZE);
    if ((request = malloc(request_size)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate memory for request", "renderVerdict");
        free(component);
        return R_MALLOC_FAIL;
    }

    snprintf(request, request_size, RENDER_VERDICT, verdict->flags, component->component_id);

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        free(component);
        return R_FAIL;
    }

    if (mysql_query(conn, request) !=0)
    {
        printf("Query failed: %s\n", request);
        printError("Failed results check in componentCheck\n", conn);
        free(request);
        free(component);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
#ifdef RZB_PERF
    pthread_mutex_lock(&perfmutex);
    perf.verdicts++;
    pthread_mutex_unlock(&perfmutex);
#endif /* PREF */
    free(request);
    free(component);
    putMYSQLConnection(conn);
    return R_SUCCESS;
}

/*
 * IN:         ALERT_TRACKER
 * OUT:        H_RESULT
 * FUNCTION:   Check the database and see if the component has already been added.
 *
*/
HRESULT componentCheck(ALERT_TRACKER *a_tracker)
{
    MYSQL *conn;
    MYSQL_RES *response;
    MYSQL_ROW row;
    ALERT_TRACKER *tracker = a_tracker;

    char *request;
    size_t request_size;
    char *protected_md5;
    unsigned row_count;

    // Make the md5sum safe for datbase usage
    if ((protectHex(tracker->blob_md5, MD5_SIZE, &protected_md5)) != R_SUCCESS)
        return R_MALLOC_FAIL;

    // Build and execute the query to check the MD5/size combo
    request_size = sizeof(COMPONENT_CHECK_SQL) + ((2*MD5_SIZE)+1) + (MAX_INT_STRING_SIZE * 2);

    if ((request = malloc(request_size)) == NULL)
        return R_MALLOC_FAIL;

    snprintf(request, request_size, COMPONENT_CHECK_SQL, protected_md5, tracker->blob_size);

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        return R_FAIL;
    }
    if (mysql_query(conn, request) !=0)
    {
        printf("Query failed: %s\n", request);
        printError("Failed results check in componentCheck", conn);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(request);

    // Check response.  If there are no rows, it isn't in the databse
    response = mysql_store_result(conn);
    row_count = mysql_num_rows(response);

    if (row_count == 0)
    {
        printf("Component not in the database!\n");
        mysql_free_result(response);
        putMYSQLConnection(conn);
        return R_NOT_FOUND;
    }

    // If it is in the database, update the alert entry.
    // If appropriate, set the block_id and track the component_id
    row = mysql_fetch_row(response);
    if (row[0] == NULL || row[1] == NULL)
    {
        printf("Returning no in db because of error\n");
        mysql_free_result(response);
        putMYSQLConnection(conn);
        return R_NOT_FOUND;
    }

    // If necessary, fetch the BlockID and populate tracker
    // Populate tracker with the appropriate componentID
    tracker->current_component_id = strtoul(row[1], NULL, 10);
    tracker->block_id = strtoul(row[0], NULL, 10);
    // Being in the database, but having no block id indicates
    // that the block was added as part of a metadata put.  The
    // data block still needs to be added.
    if (tracker->block_id == 0)
    {
        mysql_free_result(response);
        putMYSQLConnection(conn);
        return R_NOT_FOUND;
    }
    mysql_free_result(response);

    // XXX Break out to its own fuction
    // Update the appropriate field in the DB
    request_size = sizeof(UPDATE_ALERT_WITH_DATA_SQL) + tracker->blob_alert_field_len + (MAX_INT_STRING_SIZE * 3);
    if ((request = malloc(request_size)) == NULL)
    {
        putMYSQLConnection(conn);
        return R_MALLOC_FAIL;
    }
    snprintf(request, request_size, UPDATE_ALERT_WITH_DATA_SQL, tracker->blob_alert_field, tracker->current_component_id, tracker->block_id, tracker->alert_id);
    if (mysql_query(conn, request) !=0)
    {
        printf("Query failed: %s\n", request);
        printError("Failed results check in componentCheck\n", conn);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // If it is a normalized block, you also have to tie it to the non-normalized data
    // XXX Break out to its own function
    if (tracker->blob_type == NORM_BLOCK)
    {
        request_size = sizeof(UPDATE_COMPONENT_WITH_NORM_SQL) + (MAX_INT_STRING_SIZE * 2);
        if ((request = realloc(request, request_size)) == NULL)
        {
            putMYSQLConnection(conn);
            return R_MALLOC_FAIL;
        }
        snprintf(request, request_size, UPDATE_COMPONENT_WITH_NORM_SQL, tracker->current_component_id, tracker->data_component_id);

        if (mysql_query (conn, request) !=0)
        {
            printf("Query failed: %s\n", request);
            printError("Failed results check in componentCheck\n", conn);
            free(request);
            putMYSQLConnection(conn);
            return R_DB_FAIL;
        }
    }
    putMYSQLConnection(conn);
    return R_SUCCESS;
}

HRESULT insertBlob(ALERT_TRACKER *a_tracker)
{
    MYSQL *conn;
    my_bool is_null = 1;
    MYSQL_STMT *stmt = NULL;
    MYSQL_BIND bind[INSERT_BLOB_SQL_ARGC];

    ALERT_TRACKER *tracker = a_tracker;
    size_t request_size;
    char *request;

    // Still have to copy in the field name first
    request_size = sizeof(INSERT_BLOB_SQL) + strlen(tracker->blob_string);

    if ((request = malloc(request_size)) == NULL)
        return R_MALLOC_FAIL;

    snprintf(request, request_size, INSERT_BLOB_SQL, tracker->blob_string);

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        return R_FAIL;
    }
    // Init the statement
    if ((stmt = mysql_stmt_init(conn)) == NULL)
    {
        fprintf(stderr, "%s\n", mysql_stmt_error(stmt));
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // Prepare the statement
    if (mysql_stmt_prepare(stmt, request, strlen(request)))
    {
        fprintf(stderr, "Error in mysql_stmt_prepare: %s\n", mysql_stmt_error(stmt));
        mysql_stmt_close(stmt);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(request);

    // Make sure the query hasn't changed
    if (mysql_stmt_param_count(stmt) != INSERT_BLOB_SQL_ARGC)
    {
        fprintf(stderr, "%s\n", mysql_stmt_error(stmt));
        mysql_stmt_close(stmt);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // Zero out the args
    memset(bind, 0, sizeof(bind));

    // Data
    bind[0].buffer_type = MYSQL_TYPE_STRING;
    bind[0].buffer = (char *)tracker->blob_data;
    bind[0].is_null = (tracker->blob_size == 0) ? &is_null : 0;
    bind[0].length = (unsigned long *)&tracker->blob_size;

    // Alert_ID
    bind[1].buffer_type = MYSQL_TYPE_LONG;
    bind[1].buffer = (char *)&tracker->alert_id;
    bind[1].is_null = (tracker->alert_id == 0) ? &is_null : 0;
    bind[1].length = 0;
    bind[1].is_unsigned = 1;

    // bind the arguments
    if (mysql_stmt_bind_param(stmt, bind))
    {
        fprintf(stderr, "%s\n", mysql_stmt_error(stmt));
        mysql_stmt_close(stmt);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // Finally execute the statement
    if (mysql_stmt_execute(stmt))
    {
        fprintf(stderr, "%s\n", mysql_stmt_error(stmt));
        mysql_stmt_close(stmt);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // Close this statement because we are done
    mysql_stmt_close(stmt);
    putMYSQLConnection(conn);

    return R_SUCCESS;
}

/*
 * IN:            *tracker
 * OUT:           HRESULT
 * FUNCTION:      Manages the addition of new binary data into the component management
 *                portions of the database
*/
HRESULT insertComponentBlock (ALERT_TRACKER *a_tracker)
{
    MYSQL *conn;
    ALERT_TRACKER *tracker = a_tracker;
    size_t request_size;
    char *request;
    char *protected_data;
    char *protected_md5;
    char uuid_str[SIZE_OF_UUID_STRING];

    // Normalize the data for db insertion
    if ((protectHex(tracker->blob_data, tracker->blob_size, &protected_data)) != R_SUCCESS)
        return R_FAIL;
    if ((protectHex(tracker->blob_md5, MD5_SIZE, &protected_md5)) != R_SUCCESS)
        return R_FAIL;

    // Insert the component XXX Break this out as its own function
    if (tracker->current_component_id == 0)
    {
        request_size = sizeof(INSERT_NEW_COMPONENT_SQL) + ((tracker->blob_size*2)+1) + MAX_INT_STRING_SIZE*2 + ((2*MD5_SIZE) +1) + SIZE_OF_UUID_STRING;
        if (( request = malloc(request_size)) == NULL)
            return R_MALLOC_FAIL;
        uuid_unparse(tracker->data_type, uuid_str);
        snprintf(request, request_size, INSERT_NEW_COMPONENT_SQL, protected_md5, protected_data, tracker->blob_size, SIZE_OF_UUID_STRING, uuid_str);
    }
    else
    {
        request_size = sizeof(UPDATE_WITH_DATA) + ((tracker->blob_size*2)+1) + MAX_INT_STRING_SIZE;
        if ((request = malloc(request_size)) == NULL)
            return R_MALLOC_FAIL;
        snprintf(request, request_size, UPDATE_WITH_DATA, protected_data, tracker->current_component_id);
    }

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        free(protected_data);
        free(protected_md5);
        return R_FAIL;
    }
    if (mysql_query(conn, request) != 0)
    {
        printf("Request is: %s\n", request);
        free(request);
        free(protected_data);
        free(protected_md5);
        printError("Failed insert of component\n", conn);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    printf("tracker->blob_type is %u\n", tracker->blob_type);
    printf("CCID: %u, CBID: %u\n", tracker->current_component_id, tracker->block_id);


    // Populate component ID entries XXX Break this out as its own function
    if (tracker->current_component_id == 0)
        tracker->current_component_id = mysql_insert_id(conn);
    if (tracker->blob_type == FULL_BLOCK)
        tracker->data_component_id = tracker->current_component_id;

    // Protected data and request might be HUGE...so free immediately
    free(protected_data);
    free(request);

    // If necessary, insert the new component block and point the component to it
    // XXX Break out as its own function
    if (tracker->block_id == 0)
    {
        request_size = sizeof(INSERT_NEW_COMPONENT_BLOCK_SQL) + MAX_INT_STRING_SIZE;

        if ((request = malloc(request_size)) == NULL)
        {
            putMYSQLConnection(conn);
            return R_MALLOC_FAIL;
        }
        else
            memset(request, 0, request_size);

        snprintf(request, request_size, INSERT_NEW_COMPONENT_BLOCK_SQL, tracker->current_component_id);

        printf("request: %s\n", request);
        if (mysql_query(conn, request) != 0)
        {
            printError("Failed insert of component block\n", conn);
            free(request);
            putMYSQLConnection(conn);
            return R_DB_FAIL;
        }
        tracker->block_id = mysql_insert_id(conn);
        free(request);
    }
    printf("After block insert\n");
    printf("CCID: %u, CBID: %u\n", tracker->current_component_id, tracker->block_id);


    // Update component entry with block_id
    // XXX Break out as its own function
    request_size = sizeof(UPDATE_COMPONENT_WITH_BLOCK_SQL) + (MAX_INT_STRING_SIZE * 2);
    if ((request = malloc(request_size)) == NULL)
    {
        putMYSQLConnection(conn);
        return R_MALLOC_FAIL;
    }
    snprintf(request, request_size, UPDATE_COMPONENT_WITH_BLOCK_SQL, tracker->block_id, tracker->current_component_id);

    printf("request: %s\n", request);
    if (mysql_query (conn, request) != 0)
    {
        printError("Failed update of component\n", conn);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // Update the alert to point to the new component
    // XXX Break out as its own function
    request_size = sizeof(UPDATE_ALERT_WITH_DATA_SQL) + tracker->blob_alert_field_len + (MAX_INT_STRING_SIZE * 2);
    if ((request = realloc(request, request_size)) == NULL)
    {
        putMYSQLConnection(conn);
        return R_MALLOC_FAIL;
    }
    snprintf(request, request_size, UPDATE_ALERT_WITH_DATA_SQL, tracker->blob_alert_field, tracker->current_component_id, tracker->block_id, tracker->alert_id);

    printf("request: %s\n", request);
    if (mysql_query(conn, request) != 0)
    {
        printError("Failed update of component\n", conn);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(request);

    // If data block is a norm block, associate it to the primary data block
    if (tracker->blob_type == NORM_BLOCK)
        linkNormToData(tracker->current_component_id, tracker->data_component_id);

    putMYSQLConnection(conn);
    return R_SUCCESS;
}

HRESULT linkNormToData (unsigned norm_id, unsigned data_id)
{
    MYSQL *conn;
    size_t request_size;
    char *request;

    request_size = sizeof(UPDATE_COMPONENT_WITH_NORM_SQL) + (MAX_INT_STRING_SIZE*2);
    if ((request = malloc(request_size)) == NULL)
        return R_MALLOC_FAIL;

    snprintf(request, request_size, UPDATE_COMPONENT_WITH_NORM_SQL, norm_id, data_id);

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        return R_FAIL;
    }
    if (mysql_query(conn, request) !=0)
    {
        printf("Query failed: %s\n", request);
        printError("Failed results check in linkNormToData\n", conn);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(request);
    putMYSQLConnection(conn);
    return R_SUCCESS;
}

HRESULT insertAlert (ALERT_HEADER *data, unsigned msgSize, unsigned char *message, unsigned *alertID)
{
    MYSQL *conn;
    uint32_t src_ip = 0, dst_ip = 0;
    my_bool is_null = 1;
    MYSQL_STMT *stmt = NULL;
    MYSQL_BIND bind[INSERT_ALERT_SQL_ARGC];

    if ((conn = getMYSQLConnection()) == NULL)
        return R_FAIL;

    // Init the statement
    if ((stmt = mysql_stmt_init(conn)) == NULL)
    {
        printError("mysql_stmt_init() failed\n", conn);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // Prepare the statement
    if (mysql_stmt_prepare(stmt, INSERT_ALERT_SQL, sizeof(INSERT_ALERT_SQL)))
    {
        printError("mysql_stmt_prepare() failed\n", conn);
        mysql_stmt_close(stmt);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // Make sure the query hasn't changed
    if (mysql_stmt_param_count(stmt) != INSERT_ALERT_SQL_ARGC)
    {
        printError("Invalid parameter count returned by MySQL\n", conn);
        mysql_stmt_close(stmt);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // Zero out the args
    memset(bind, 0, sizeof(bind));

    // data->event_id
    bind[0].buffer_type = MYSQL_TYPE_LONG;
    bind[0].buffer = (char *)&data->event_id;
    bind[0].is_null = (data->event_id == 0) ? &is_null : 0;
    bind[0].length = 0;
    bind[0].is_unsigned = 1;

    // data->nugget_id
    bind[1].buffer_type = MYSQL_TYPE_LONG;
    bind[1].buffer = (char *)&data->nugget_id;
    bind[1].is_null = (data->nugget_id == 0) ? &is_null : 0;
    bind[1].length = 0;
    bind[1].is_unsigned = 1;

    // data->timestamp
    bind[2].buffer_type = MYSQL_TYPE_LONG;
    bind[2].buffer = (char *)&data->timestamp;
    bind[2].is_null = (data->timestamp == 0) ? &is_null : 0;
    bind[2].length = 0;

    // data->priority
    bind[3].buffer_type = MYSQL_TYPE_TINY;
    bind[3].buffer = (char *)&data->priority;
    bind[3].is_null = (data->priority == 0) ? &is_null : 0;
    bind[3].length = 0;

    // message
    bind[4].buffer_type = MYSQL_TYPE_STRING;
    bind[4].buffer = (char *)message;
    bind[4].is_null = (msgSize == 0) ? &is_null : 0;
    bind[4].length = (unsigned long *)&msgSize;

    // src_ip
    src_ip = htonl(data->src_ip);
    bind[5].buffer_type = MYSQL_TYPE_LONG;
    bind[5].buffer = (char *)&src_ip;
    bind[5].is_null = (src_ip == 0) ? &is_null : 0;
    bind[5].is_unsigned = 1;

    // dst_ip
    dst_ip = htonl(data->dst_ip);
    bind[6].buffer_type = MYSQL_TYPE_LONG;
    bind[6].buffer = (char *)&dst_ip;
    bind[6].is_null = (dst_ip == 0) ? &is_null : 0;
    bind[6].is_unsigned = 1;

    // ip_proto
    bind[7].buffer_type = MYSQL_TYPE_SHORT;
    bind[7].buffer = (char *)&(data->ip_proto);
    bind[7].is_null = (data->ip_proto == 0) ? &is_null : 0;
    bind[7].length = 0;
    bind[7].is_unsigned = 1;

    // src_port
    bind[8].buffer_type = MYSQL_TYPE_SHORT;
    bind[8].buffer = (char *)&(data->src_port);
    bind[8].is_null = (data->src_port == 0) ? &is_null : 0;
    bind[8].length = 0;
    bind[8].is_unsigned = 1;

    // dst_port
    bind[9].buffer_type = MYSQL_TYPE_SHORT;
    bind[9].buffer = (char *)&(data->dst_port);
    bind[9].is_null = (data->dst_port == 0) ? &is_null : 0;
    bind[9].length = 0;
    bind[9].is_unsigned = 1;

    // bind the arguments
    if (mysql_stmt_bind_param(stmt, bind))
    {
        perror("mysql_stmt_bind_param() failed");
        mysql_stmt_close(stmt);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // Finally execute the statement
    if (mysql_stmt_execute(stmt))
    {
        fprintf(stderr, "%s\n", mysql_stmt_error(stmt));
        mysql_stmt_close(stmt);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // Close this statement because we are done
    mysql_stmt_close(stmt);

    // Fetch the alertID
    if ((*alertID = mysql_insert_id(conn)) == 0)
    {
        printError("Failed to retrieve new alert_ID\n", conn);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    data->alert_id = *alertID;
#ifdef RZB_PERF
    pthread_mutex_lock(&perfmutex);
    perf.alerts++;
    pthread_mutex_unlock(&perfmutex);
#endif /* RZB_PERF */

    putMYSQLConnection(conn);
    return R_SUCCESS;
}

