#include <uuid/uuid.h>
#include <mysql.h>
#include <syslog.h>

#include "config.h"
#include "rzb_global.h"
#include "rzb_conf.h"
#include "rzb_client.h"
#include "rzb_db.h"
#include "rzb_network.h"
#include "rzb_dispatcher.h"
#include "rzb_nug_manage.h"
#include "rzb_alert_listener.h"
#include "rzb_alert_global.h"
#include "rzb_alert_db.h"
#include "rzb_component_db.h"
#include "rzb_management_log.h"

static HRESULT requestBlob(ALERT_TRACKER *a_tracker);
static HRESULT sendAlertNotify(const char *ip, const char *port, ALERT_HEADER *alert, SNAKECHARMER *scpacket);

static HRESULT alertNotify(ALERT_HEADER *alert)
{
    SNAKECHARMER scpacket;

    if (alert == NULL || (nrtconfig.on1_ip == NULL && nrtconfig.on2_ip == NULL))
    {
        submitLog(LOG_ERR, "Unable to send alert notifications", "alertNotify");
        return R_FAIL;
    }

    protocolFill(&scpacket);
    scpacket.msgtype = ALRT;

    if (nrtconfig.on1_ip != NULL)
        sendAlertNotify(nrtconfig.on1_ip, nrtconfig.on1_port, alert, &scpacket);

    if (nrtconfig.on2_ip != NULL)
        sendAlertNotify(nrtconfig.on2_ip, nrtconfig.on2_port, alert, &scpacket);

    return R_SUCCESS;
}

static HRESULT sendAlertNotify(const char *ip, const char *port, ALERT_HEADER *alert, SNAKECHARMER *scpacket)
{
    HRESULT rval;
    RZBNetworkConnection *net = ConnectMe(ip, port);
    if (!net)
    {
        printf("connection to %s:%s failed\n", ip, port);
        return R_FAIL;
    }

    rval = R_FAIL;
    do
    {
        if (sendWrap(net, scpacket, sizeof(*scpacket)) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to send scpacket", "sendAlertNotify");
            break;
        }

        if (sendWrap(net, alert, sizeof(*alert)) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to send alert header packet", "sendAlertNotify");
            break;
        }

        if (recvWrap(net, scpacket, sizeof(*scpacket)) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to receive scpacket", "sendAlertNotify");
            break;
        }

        if (scpacket->msgtype == ACK)
            rval = R_SUCCESS;

    } while (0);

    NetworkConnectionDestroy(net);
    return rval;
}

HRESULT handleAlertBlob(RZBNetworkConnection *net)
{
    BLOB_REQUEST requestpkt;
    BLOB_HEADER headerpkt;
    SNAKECHARMER scpacket;
    unsigned char *data;
    unsigned size;

    if (recvWrap(net, &requestpkt, sizeof(requestpkt)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to receive blob request", "handleAlertBlob");
        return R_FAIL;
    }

    if (getDataByAlertID(requestpkt.blob_type, &data, &size, requestpkt.alertID) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to get data", "handleAlertBlob");
        return R_FAIL;
    }

    if (!size)
    {
        submitLog(LOG_ERR, "Data returned as 0 size", "handleAlertBlob");
        return R_FAIL;
    }

    headerpkt.alertID = requestpkt.alertID;
    headerpkt.blobType = requestpkt.blob_type;
    headerpkt.blobSize = size;

    if (sendWrap(net, &headerpkt, sizeof(headerpkt)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send header packet", "sendAlertNotify");
        free(data);
        return R_FAIL;
    }

    if (sendWrap(net, data, size) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send data", "sendAlertNotify");
        free(data);
        return R_FAIL;
    }

    free(data);

    if (recvWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to receive scpacket", "sendAlertNotify");
        return R_FAIL;
    }

    if (scpacket.msgtype != ACK)
    {
        submitLog(LOG_ERR, "Did not receive an ACK packet", "sendAlertNotify");
        return R_FAIL;
    }

    return R_SUCCESS;
}

HRESULT handleJudgement(RZBNetworkConnection *net)
{
    JUDGEMENT verdict;
    SNAKECHARMER scpacket;

    if (recvWrap(net, &verdict, sizeof(verdict)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to receive judgement", "handleVerdict");
        sendFailure(net, &scpacket);
    }

    if (renderVerdict(&verdict) == R_FAIL)
    {
        submitLog(LOG_ERR, "Failed to render verdict", "handleVerdict");
        sendFailure(net, &scpacket);
        return R_FAIL;
    }
    acknowledge(net, &scpacket);
    return R_SUCCESS;
}

HRESULT handleAlert(RZBNetworkConnection *net)
{
    SNAKECHARMER scpacket;
    ALERT_HEADER data;
    ALERT_TRACKER tracker;
    unsigned msgSize;
    unsigned char *message;
    unsigned realid = 0;

    // Capture initial alert information
    if (recvWrap(net, &data, sizeof(data)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to get aleret header", "handleAlert");
        return R_FAIL;
    }

    if (recvWrap(net, &msgSize, sizeof(msgSize)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to get message size", "handleAlert");
        return R_FAIL;
    }

    if ((message = malloc(msgSize)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate space for message", "handleAlert");
        return R_FAIL;
    }

    if (recvWrap(net, message, msgSize) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to receive message", "handleAlert");
        free(message);
        return R_FAIL;
    }

    /* Translate event_id */
    if (ResolveEventLookup(data.event_id, &realid) == R_NOT_FOUND)
    {
        submitLog(LOG_ERR, "Temp event id not in table", "handleAlert");
        sendFailure(net, &scpacket);
        free(message);
        return R_FAIL;
    }
    else if (!realid)
    { // If the id is still 0, there was no resolution
        submitLog(LOG_ERR, "Resolving and setting event ID", "handleAlert");
        getEventID(&realid);
        ResolveEventLookup(data.event_id, &realid); // This sets the event id
    }
    data.event_id = realid;
    /* event_id now matches the id in the db */

    if (insertAlert(&data, msgSize, message, &tracker.alert_id) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "insertAlert failed", "handleAlert");
        free(message);
        sendFailure(net, &scpacket);
        return R_FAIL;
    }

    // Finished with message and data
    free(message);

    if (!data.flags1)
    {
        submitLog(LOG_ERR, "insertAlert failed", "handleAlert");
        acknowledge(net, &scpacket);
        return R_SUCCESS;
    }

    acknowledge(net, &scpacket);

    // Prep global values in tracker
    tracker.main_md5 = data.main_md5;
    tracker.net = net;
    tracker.blob_md5 = NULL;
    tracker.blob_data = NULL;
    tracker.block_id = 0;
    tracker.data_component_id = 0;
    tracker.current_component_id = 0;
    uuid_copy(tracker.data_type, data.dataType);

    // Fetch additional data
    if (data.flags1 & SHORT_DATA)
    {
        tracker.blob_type = SHORT_DATA;
        tracker.blob_string = SHORT_DATA_STRING;
        tracker.blob_string_len = sizeof(SHORT_DATA_STRING);
        if ((requestBlob(&tracker)) != R_SUCCESS)
        {
            sendFailure(net, &scpacket);
            submitLog(LOG_ERR, "Failed to retrieve short data", "handleAlert");
            return R_FAIL;
        }
        free(tracker.blob_data);
        tracker.blob_data = NULL;
    }
    if (data.flags1 & LONG_DATA)
    {
        tracker.blob_type = LONG_DATA;
        tracker.blob_string = LONG_DATA_STRING;
        tracker.blob_string_len = sizeof(LONG_DATA_STRING);
        if ((requestBlob(&tracker)) != R_SUCCESS)
        {
            sendFailure(net, &scpacket);
            submitLog(LOG_ERR, "Failed to retrieve long data", "handleAlert");
            return R_FAIL;
        }
        free(tracker.blob_data);
        tracker.blob_data = NULL;
    }

    // Data blocks require a component check
    if (data.flags1 & FULL_BLOCK)
    {
        tracker.blob_type = FULL_BLOCK;
        tracker.blob_string = FULL_BLOCK_STRING;
        tracker.blob_string_len = sizeof(FULL_BLOCK_STRING);
        tracker.blob_md5 = data.db_md5;
        tracker.blob_size = data.db_size;
        tracker.blob_alert_field = FULL_ALERT_FIELD;
        tracker.blob_alert_field_len = sizeof(FULL_ALERT_FIELD) ;

        if ((componentCheck(&tracker)) != R_SUCCESS)
        {
            if ((requestBlob(&tracker)) != R_SUCCESS)
            {
                sendFailure(net, &scpacket);
                submitLog(LOG_ERR, "Failed to retrieve full block", "handleAlert");
                return R_FAIL;
            }
        }
        free(tracker.blob_data);
        tracker.blob_data = NULL;
    }

    tracker.current_component_id = 0;
    if (data.flags1 & NORM_BLOCK)
    {
        tracker.blob_type = NORM_BLOCK;
        tracker.blob_string = NORM_BLOCK_STRING;
        tracker.blob_string_len = sizeof(NORM_BLOCK_STRING);
        tracker.blob_md5 = data.norm_md5;
        tracker.blob_size = data.norm_size;
        tracker.blob_alert_field = NORM_ALERT_FIELD;
        tracker.blob_alert_field_len = sizeof(NORM_ALERT_FIELD);

        if ((componentCheck(&tracker)) != R_SUCCESS)
        {
            if ((requestBlob(&tracker)) != R_SUCCESS)
            {
                submitLog(LOG_ERR, "Failed to retrieve norm block", "handleAlert");
                sendFailure(net, &scpacket);
                return R_FAIL;
            }
        }
        free(tracker.blob_data);
        tracker.blob_data = NULL;
    }

    sendFailure(net, &scpacket); // XXX This was a RST packet before sendFailue() was added
                                 // Wouldn't it be more appropriate here to response with an ACK?
                                 //
    if (nrtconfig.on1_ip != NULL || nrtconfig.on2_ip != NULL)
        alertNotify(&data);

    return R_SUCCESS;
}

static HRESULT requestBlob(ALERT_TRACKER *a_tracker)
{
    SNAKECHARMER scpacket;
    BLOB_REQUEST blobRequest;
    BLOB_HEADER blobResponse;
    ALERT_TRACKER *tracker = a_tracker;

    protocolFill(&scpacket);
    scpacket.msgtype = REQBLOB;

    blobRequest.alertID = tracker->alert_id;
    blobRequest.blob_type = tracker->blob_type;
    tracker->blob_size = 0;

    // Prepare to grab the data
    if (sendWrap(tracker->net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send REQBLOB packet", "requestBlob");
        return R_FAIL;
    }
    if (sendWrap(tracker->net, &blobRequest, sizeof(blobRequest)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send BLOB_REQUEST packet", "requestBlob");
        return R_FAIL;
    }
    if (recvWrap(tracker->net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to receive response SNAKECHARMER packet", "requestBlob");
        return R_FAIL;
    }
    if (scpacket.msgtype == SENDBLOB)
    {
        if (recvWrap(tracker->net, &blobResponse, sizeof(blobResponse)) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to receive blob response (BLOB_HEADER)", "requestBlob");
            return R_FAIL;
        }
    }
    else
    {
        submitLog(LOG_ERR, "Invalid response type (not SENDBLOB)", "requestBlob");
        return R_FAIL;
    }

    tracker->blob_size = blobResponse.blobSize;

    // If we were succesful, blob_size should be populated
    if (!tracker->blob_size)
    {
        submitLog(LOG_ERR, "Blob size is 0 after insert", "requestBlob");
        return R_FAIL;
    }

    if ((tracker->blob_data = malloc(tracker->blob_size)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to alloc memoery for blob_data", "requestBlob");
        return R_FAIL;
    }

    // Grab Data Block
    if (recvWrap(tracker->net, tracker->blob_data, tracker->blob_size) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to received blob_data", "requestBlob");
        return R_FAIL;
    }

    acknowledge(tracker->net, &scpacket);

    if (tracker->blob_type == FULL_BLOCK || tracker->blob_type == NORM_BLOCK)
    {
        if ((insertComponentBlock(tracker)) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to insert component block", "requestBlob");
            return R_FAIL;
        }
    }
    else if ((insertBlob(tracker)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to insert blob", "requestBlob");
        return R_FAIL;
    }
    return R_SUCCESS;
}

