#include <mysql.h>
#include <string.h>
#include <syslog.h>

#include "config.h"
#include "rzb_global.h"
#include "rzb_component_db.h"
#include "rzb_alert_global.h"
#include "rzb_db.h"
#include "rzb_management_log.h"
#include "rzb_conf.h"
#include "rzb_client.h"

// SQL STATEMENTS
static const char GET_COMPONENT[] = "SELECT Component_ID, Norm_ID, Datatype, Threat_Flags, Custom_Flags FROM components WHERE (Checksum = '%s' AND Size = %u)";
//static const char GET_COMPONENT_ID[] = "SELECT Component_ID FROM components WHERE (Checksum = '%s' AND Size = %u)";
//static const char GET_NORMAL_ID[] = "SELECT Norm_ID FROM components WHERE (Checksum = '%s' AND Size = %u)";
//static const char INSERT_DATA[] = "UPDATE components SET data=(unhex('%s')) WHERE Component_ID=%u";

HRESULT insertBaseComponent(BASIC_COMPONENT *component)
{
    static const char INSERT_BASE[] = "INSERT INTO components (Checksum, Size, Datatype) VALUES ('%s', %u, '%s')";
    MYSQL *conn;
    char *safe_md5;
    char *request;
    size_t request_size;
    char uuid_string[SIZE_OF_UUID_STRING];

    /* NORMALIZE MD5 */
    if ((protectHex(component->md5, MD5_SIZE, &safe_md5)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to protect md5 hex", "insertBasicComponent");
        return R_FAIL;
    }

    /* BUILD QUERY */
    request_size = sizeof(INSERT_BASE) + strlen(safe_md5) + MAX_INT_STRING_SIZE + MAX_DESCRIPTION_SIZE;
    if ((request = malloc(request_size)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate momry for request", "insertBasicComponent");
        free(safe_md5);
        return R_MALLOC_FAIL;
    }

    uuid_unparse(component->datatype, uuid_string);
    snprintf(request, request_size, INSERT_BASE, safe_md5, component->size, uuid_string);

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        free(safe_md5);
        return R_FAIL;
    }

    /* EXECUTE QUERY */
    if (mysql_query(conn, request) != 0)
    {
        printf("Request: %s\n", request);
        printError("Failed ot insert block\n", conn);
        free(request);
        free(safe_md5);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    /* GRAB COMPONENT_ID */
    component->component_id = mysql_insert_id(conn);
    free(request);
    free(safe_md5);
    putMYSQLConnection(conn);
    return R_SUCCESS;
}

/* makeRequestSafe
 * Rebuild a string to be mysql safe.
 *
 * IN:      unsigned char request (passed by reference), unsigned size
 * OUT:     HRESULT, Set request pointer to new buffer
 * NOTE:    This function assigns a new value to the request pointer and
 *          then frees the old pointer
*/
static HRESULT makeRequestSafe(char **request, unsigned size)
{
    MYSQL *conn;
    char *newRequest = malloc(size*2+1);
    void *temp;

    if (newRequest == NULL)
        return R_MALLOC_FAIL;

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(newRequest);
        return R_FAIL;
    }
    mysql_real_escape_string(conn, newRequest, *request, size);
    temp = *request;
    *request = newRequest;
    free(temp);
    putMYSQLConnection(conn);
    return R_SUCCESS;
}

HRESULT checkComponent(const unsigned char *md5, unsigned size, COMPONENT *component)
{
    static const char CHECK_COMPONENT[] = "SELECT Component_ID FROM components WHERE (checksum = '%s' AND Size = %u)";
    MYSQL *conn;
    MYSQL_RES *response;
    MYSQL_ROW row;

    char request[200];
    char *safe_md5;

    if ((protectHex(md5, MD5_SIZE, &safe_md5)) != R_SUCCESS)
        return R_MALLOC_FAIL;

    snprintf(request, sizeof(request), CHECK_COMPONENT, safe_md5, size);
    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(safe_md5);
        return R_FAIL;
    }
    if (mysql_query(conn, request) !=0)
    {
        printf("Query failed: %s\n", request);
        printError("Failed results check in checkComponent\n", conn);
        free(safe_md5);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(safe_md5);

    response = mysql_store_result(conn);
    if (!response)
    {
        if (mysql_field_count(conn) == 0)
        {
            putMYSQLConnection(conn);
            return R_NOT_FOUND;
        }
        else
        {
            printError("Failed to get result set ofr getComponent\n", conn);
            putMYSQLConnection(conn);
            return R_DB_FAIL;
        }
    }
    if ((row = mysql_fetch_row(response)) == NULL)
    {
        mysql_free_result(response);
        putMYSQLConnection(conn);
        return R_FAIL;
    }

    // If the component is there, populate component structure
    if (component != NULL)
        component->component_id = strtoul(row[0], NULL, 10);
    mysql_free_result(response);
    putMYSQLConnection(conn);
    return R_SUCCESS;
}

/* getComponent
 * Retrieve a component by MD5, and if it exists, populate a new component structure
 *
 * IN:      *md5, unsigned size, COMPONENT (passed by reference)
 * OUT:     HRESULT and populates the COMPONENT pointer.
 * NOTE:    If the MD5/Size combination does not exist, then COMPONENT is NULL and the
 *          function returns R_NOT_FOUND.  If you are not going to work with the data
 *          returned, and only need to know if the entry is in the database, then use
 *          checkMD5 to avoid the overhead of copying the data to a structure.
*/
HRESULT getComponent(const unsigned char *md5, unsigned size, COMPONENT *component)
{
    MYSQL *conn;
    MYSQL_RES *response;
    MYSQL_ROW row;
    char *request;
    size_t request_size;
    char *safe_md5;

    // Build safe request
    request_size = sizeof(GET_COMPONENT) + SAFE_MD5_SIZE + MAX_INT_STRING_SIZE;
    if ((protectHex(md5, MD5_SIZE, &safe_md5)) != R_SUCCESS)
        return R_MALLOC_FAIL;
    if ((request = malloc(request_size)) == NULL)         // Malloc request
        return R_MALLOC_FAIL;
    snprintf(request, request_size, GET_COMPONENT, safe_md5, size);

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(safe_md5);
        return R_FAIL;
    }
    // Execute query
    if (mysql_query(conn, request) !=0)
    {
        printf("Query failed: %s\n", request);
        printError("Failed results check in getComponent\n", conn);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(request);

    // Check response
    response = mysql_store_result(conn);
    if (!response)
    {
        if (mysql_field_count(conn) == 0)
            return R_NOT_FOUND;
        printError("Failed to get result set ofr getComponent\n", conn);
        return R_DB_FAIL;
    }

    if ((row = mysql_fetch_row(response)) == NULL)
    {
        putMYSQLConnection(conn);
        return R_NOT_FOUND;
    }
    if (row[0] == NULL)
    {
        printError("Row 0 is null in getComponent\n", conn);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // If the component is there, populate component structure
    component->component_id = strtoul(row[0], NULL, 10);
    component->norm_id = strtoul(row[1], NULL, 10);
    memcpy(component->md5, md5, MD5_SIZE);
    component->size = size;
    component->data = NULL;                 // This function does not grab the data block
    component->norm = NULL;                 // This function does not grab the normalized data
    component->sc_list = NULL;              // This function does not populate the sub-component list
    mysql_free_result(response);
    putMYSQLConnection(conn);

    return R_SUCCESS;
}

HRESULT doesComponentExist(const unsigned char *md5, unsigned size)
{
    MYSQL *conn;
    MYSQL_RES *response;
    char  *request;
    size_t request_size;
    char *safe_md5;

    // Build safe request
    request_size = sizeof(GET_COMPONENT) + SAFE_MD5_SIZE + MAX_INT_STRING_SIZE;
    if ((protectHex(md5, MD5_SIZE, &safe_md5)) != R_SUCCESS)
        return R_MALLOC_FAIL;
    if ((request = malloc(request_size)) == NULL)         // Malloc request
        return R_MALLOC_FAIL;
    snprintf(request, request_size, GET_COMPONENT, safe_md5, size);
    free(safe_md5);
    makeRequestSafe(&request, request_size);

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        return R_FAIL;
    }
    // Execute query
    if (mysql_query(conn, request) !=0)
    {
        printf("Query failed: %s\n", request);
        printError("Failed results check in checkMD5\n", conn);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(request);                                        // Free request

    // Check response
    if ((response = mysql_store_result(conn)))
    {
        putMYSQLConnection(conn);
        return R_SUCCESS;
    }

    mysql_free_result(response);
    putMYSQLConnection(conn);
    return R_NOT_FOUND;
}

/* getComponentById
 * Retrieve a component by MD5, and if it exists, populate a new component structure
 *
 * IN:      *md5, unsigned size, COMPONENT (passed by reference)
 * OUT:     HRESULT and populates the COMPONENT pointer.
 * NOTE:    If the MD5/Size combination does not exist, then COMPONENT is NULL and the
 *          function returns R_NOT_FOUND.  If you are not going to work with the data
 *          returned, and only need to know if the entry is in the database, then use
 *          checkMD5 to avoid the overhead of copying the data to a structure.
*/
HRESULT getComponentById(unsigned id, COMPONENT **component)
{
    static const char GET_COMPONENT_BY_ID[] = "SELECT Component_ID, Norm_ID, Datatype, Threat_Flags, Custom_Flags, Checksum, Size FROM components WHERE Component_ID = %u";
    MYSQL *conn;
    MYSQL_RES      *response;
    MYSQL_ROW      row;
    COMPONENT      *localComponent = NULL;
    char  *request;
    size_t   request_size;

    // Build safe request
    request_size = strlen(GET_COMPONENT_BY_ID) + sizeof(int);
    if ((request = malloc(request_size)) == NULL)         // Malloc request
        return R_MALLOC_FAIL;
    snprintf(request, request_size, GET_COMPONENT_BY_ID, id);
    makeRequestSafe(&request, strlen(request));

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        return R_FAIL;
    }
    // Execute query
    if (mysql_query(conn, request) !=0)
    {
        printf("Query failed: %s\n", request);
        printError("Failed results check in checkMD5\n", conn);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(request);                                        // Free request

    // Check response (if conn = 0, then there was no entry matching the request)
    if (!(response = mysql_store_result(conn)))
    {
        printError("Failed to get result set for getComponent\n", conn);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    if (!mysql_field_count(conn))
    {
        putMYSQLConnection(conn);
        return R_NOT_FOUND;
    }

    if ((row = mysql_fetch_row(response)) == NULL)
    {
        mysql_free_result(response);
        putMYSQLConnection(conn);
        return R_NOT_FOUND;
    }

    if (row[0] == NULL)
    {
        printError("Row 0 is null in getComponent\n", conn);
        mysql_free_result(response);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    // If the component is there, populate component structure
    if ((localComponent = malloc(sizeof(COMPONENT))) == NULL)
    {
        mysql_free_result(response);
        putMYSQLConnection(conn);
        return R_MALLOC_FAIL;
    }

    localComponent->component_id = strtoul(row[0], NULL, 10);
    localComponent->norm_id = strtoul(row[1], NULL, 10);
    memcpy(localComponent->md5, row[5], SAFE_MD5_SIZE);
    localComponent->size = strtoul(row[6], NULL, 10);
    memcpy(localComponent->datatype, row[2],SIZE_OF_UUID_STRING);
    if (row[3] == NULL)
        localComponent->threat_flags = 0;
    else
        localComponent->threat_flags = strtoul(row[3], NULL, 10);
    if (row[4] == NULL)
        localComponent->custom_flags = 0;
    else
        localComponent->custom_flags = strtoul(row[4], NULL, 10);
    localComponent->data = NULL;                 // This function does not grab the data block
    localComponent->norm = NULL;                 // This function does not grab the normalized data
    localComponent->sc_list = NULL;              // This function does not populate the sub-component list

    mysql_free_result(response);
    putMYSQLConnection(conn);
    *component = localComponent;

    return R_SUCCESS;
}

#ifdef UNUSED_FUNCTION
HRESULT getDataBlock(COMPONENT **component, unsigned block)
{
    static const char GET_DATA_BY_ID[] = "SELECT Data FROM components WHERE Component_ID = %u";
    MYSQL *conn;
    MYSQL_RES      *response;
    MYSQL_ROW      row;
    COMPONENT      *localComponent = *component;
    char  *request;
    size_t   request_size;
    unsigned long   *lengths;

    // Build safe request
    request_size = strlen(GET_DATA_BY_ID) + sizeof(int) + 1;
    if ((request = malloc(request_size)) == NULL)         // Malloc request
        return R_MALLOC_FAIL;
    if (block == FULL_BLOCK)
        snprintf(request, request_size, GET_DATA_BY_ID, localComponent->component_id);
    else
        snprintf(request, request_size, GET_DATA_BY_ID, localComponent->norm_id);

    makeRequestSafe(&request, strlen(request));

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        return R_FAIL;
    }
    // Execute query
    if (mysql_query(conn, request) !=0)
    {
        printf("Query failed: %s\n", request);
        printError("Failed results check in checkMD5\n", conn);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(request);                                        // Free request

    // Check response (if conn = 0, then there was no entry matching the request)
    if (!(response = mysql_store_result(conn)))
    {
        printError("Failed to get result set for getComponent\n", conn);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    if (mysql_field_count(conn) == 0)
    {
        putMYSQLConnection(conn);
        return R_NOT_FOUND;
    }

    if ((row = mysql_fetch_row(response)) == NULL || row[0] == NULL)
    {
        printError("Row 0 is null in getComponent\n", conn);
        putMYSQLConnection(conn);
        return R_NOT_FOUND;
    }

    lengths = mysql_fetch_lengths(response);
    if (block == FULL_BLOCK)
    {
        localComponent->data = malloc(lengths[0]);
        memcpy(localComponent->data, row[0], lengths[0]);
    }
    else
    {
        localComponent->norm = malloc(lengths[0]);
        memcpy(localComponent->norm, row[0], lengths[0]);
    }

    mysql_free_result(response);
    putMYSQLConnection(conn);
    return R_SUCCESS;
}

HRESULT insertBaseBlock(const unsigned char *md5, unsigned size, unsigned *id)
{
    static const char INSERT_BASE_BLOCK[] = "INSERT INTO components (Checksum, Size) VALUES ('%s', %u)";
    MYSQL *conn;
    size_t request_size;
    char *request;
    char *protected_md5;

    if ((protectHex(md5, MD5_SIZE, &protected_md5)) != R_SUCCESS)
        return R_FAIL;

    request_size = strlen(INSERT_BASE_BLOCK) + sizeof(int) + SAFE_MD5_SIZE;
    if ((request = malloc(request_size)) == NULL)
    {
        free(protected_md5);
        return R_MALLOC_FAIL;
    }
    snprintf(request, request_size, INSERT_BASE_BLOCK, protected_md5, size);

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        free(protected_md5);
        return R_FAIL;
    }
    if (mysql_query(conn, request) != 0)
    {
        printf("Request: %s\n", request);
        printError("Failed ot insert block\n", conn);
        free(request);
        free(protected_md5);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(request);
    free(protected_md5);
    putMYSQLConnection(conn);
    return R_SUCCESS;
}
#endif

