#include <uuid/uuid.h>
#include <mysql.h>
#include <syslog.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/socket.h>

#include "config.h"
#include "rzb_global.h"
#include "rzb_api.h"
#include "rzb_db.h"
#include "rzb_client.h"
#include "rzb_network.h"
#include "rzb_thread.h"
#include "rzb_dispatcher.h"
#include "rzb_nug_manage.h"
#include "rzb_router.h"
#include "rzb_thread.h"
#include "rzb_alert_global.h"
#include "rzb_component_global.h"
#include "rzb_alert_db.h"
#include "rzb_component_db.h"
#include "rzb_workstation_db.h"
#include "rzb_intel_db.h"
#include "rzb_conf.h"
#include "rzb_alert_listener.h"
#include "rzb_workstation_listener.h"
#include "rzb_intel_listener.h"
#include "rzb_management_log.h"

typedef struct _EVENTLOOKUPTABLE
{
    unsigned tempid;
    unsigned realid;
    unsigned outstanding;
    struct _EVENTLOOKUPTABLE *next;
} EVENTLOOKUPTABLE;

/* Variables to handle valid registration checks */
unsigned validNuggetCount; /* Number of valid nugget types */
uuid_t *validNuggetSet; /* List of valid UUIDs in full form */
unsigned dataTypesCount;
uuid_t *dataTypes;
unsigned g_sock;

static pthread_mutex_t eventidmutex;
static pthread_mutex_t eventlookupmutex;

static EVENTLOOKUPTABLE *lookuptable = NULL;

static HRESULT handleMD5CacheCheck(RZBNetworkConnection *net);
static HRESULT FindEventLookup(unsigned tempid, EVENTLOOKUPTABLE **lookup);

#ifdef RZB_PERF
pthread_mutex_t perfmutex = PTHREAD_MUTEX_INITIALIZER;
PERF_TRACK perf;  // Global for tracking system statistics

static void initPerf(void)
{
    memset(&perf, 0, sizeof(perf));
}

static void printPerf(void)
{
    printf("\nDispatcher Statistics\n");
    printf("=====================\n\n");
    printf("DB Cache Checks:         %u\n", perf.db_cache_checks);
    printf("DB Cache Misses:         %u\n", perf.db_cache_miss);
    printf("DB Cache Hits, Good:     %u\n", perf.db_cache_hit_good);
    printf("DB Cache Hits, Bad:      %u\n", perf.db_cache_hit_bad);
    printf("DB Web Track Inserts:    %u\n", perf.web_tracker_insert);
    printf("DB Email Track Inserts:  %u\n", perf.smtp_tracker_insert);
    printf("DB DNS Track Inserts:    %u\n", perf.dns_tracker_insert);
    printf("Bytes Received:          %lu\n", perf.bytes_in);
    printf("Bytes Sent:              %lu\n", perf.bytes_out);
    printf("Database Connections:    %u\n", perf.db_conn);
    printf("\nEnd Stats\n");
    printf("=====================\n\n");
}
#endif /* RZB_PERF */

static HRESULT handleMeta(RZBNetworkConnection *net)
{
    SNAKECHARMER scpacket;
    METAPACKET mpacket;
    unsigned char *data;

    if (recvWrap(net, &mpacket, sizeof(mpacket)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Unable to receive metapacket", "handleMeta");
        sendFailure(net, &scpacket);
        return R_FAIL;
    }

    switch (mpacket.type)
    {
        case CAP_DETAILS:
            break;
        default:
            submitLog(LOG_ERR, "Invaild metadata type", "handleMeta");
            sendFailure(net, &scpacket);
            return R_FAIL;
    }

    if ((data = malloc(mpacket.size)) == NULL)
    {
        submitLog(LOG_ERR, "Unable to allocate space for data", "handleMeta");
        sendFailure(net, &scpacket);
        return R_MALLOC_FAIL;
    }

    if (recvWrap(net, data, mpacket.size) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Unable to recieve data", "handleMeta");
        sendFailure(net, &scpacket);
        free(data);
        return R_FAIL;
    }

    acknowledge(net, &scpacket);

    if (insertComponentMetaData(&mpacket, data, mpacket.size) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to insert metadata", "handleMeta");
        free(data);
        return R_FAIL;
    }

    free(data);
    return R_SUCCESS;
}

static HRESULT handleRegistration(RZBNetworkConnection *net, unsigned *nugget_id)
{
    REGPACKET regpacket;
    NUGGETENTRY *entry;
    char tmp[INET6_ADDRSTRLEN];

    if ((entry = calloc(1, sizeof(*entry))) == NULL)
    {
        printf("%s", "Failed to allocate a nugget entry\n");
        return R_MALLOC_FAIL;
    }

    if (recvWrap(net, &regpacket, sizeof(regpacket)) != R_SUCCESS)
    {
        perror("Could not get registration information");
        return R_FAIL;
    }

    entry->nugaddr = regpacket.nugaddr;
    uuid_copy(entry->nugtype, regpacket.nugtype);
    uuid_copy(entry->nugapp, regpacket.nugapp);
    uuid_copy(entry->datatype, regpacket.datatype);
    entry->nugport = regpacket.nugport;
    entry->freeinstances = regpacket.freeinstances;
    entry->nuggetid = regpacket.nuggetid;

    snprintf(entry->name, sizeof(entry->name), "%s", regpacket.name);

    /* Register nugget with database */
    dbRegisterNugget(entry);

    /* If we have a new non-0 registration, ensure the number does not already exist in the routing table */
    if (entry->nuggetid)
    {
        if (!uuid_compare(entry->nugtype, DETECTION))
            removeRouteByNugID(entry->nuggetid);
        deleteNuggetEntry(entry->nuggetid);
    }

    /* Now pass to the system that manages in-memory handling of nugget information */
    if ((addNuggetEntry(entry)) == R_SUCCESS)
    {
        tmp[0] = 0;
        printf("\nNugget Registration: %s:%s:%d\n",
               entry->name, inet_ntop(AF_INET, &entry->nugaddr, tmp, sizeof(tmp)), entry->nugport);
        *nugget_id = entry->nuggetid;
        printRoutingTable();
        //printTypeTable();
        //printDataTable();
        return R_SUCCESS;
    }
    else
    {
        printf("Nugget Registration failed\n");
        return R_FAIL;
    }
}

static HRESULT handleNewCapability(RZBNetworkConnection *net)
{
    NEWCAPPACKET newcap;
    NUGGETENTRY *target = NULL;
    uuid_t type;
    HRESULT result;

    if (recvWrap(net, &newcap, sizeof(newcap)) != R_SUCCESS)
    {
        perror("Could not get update information from new cap packet\n");
        return R_FAIL;
    }

    uuid_copy(type, newcap.nugtype);

    result = findByID(&target, newcap.nuggetid, type);

    if (result == R_NOT_FOUND)
    {
        printf("Failed to find nugget to update\n");
        return R_FAIL;
    }
    else if (result == R_FAIL)
    {
        printf("Not sure whats wrong, but handleNewCapability failed to get id\n");
        return R_FAIL;
    }

    if (newcap.nugport == target->nugport && newcap.nugaddr.s_addr == target->nugaddr.s_addr)
    {
        /* Data type is the same */
        if ((uuid_compare(newcap.datatype, target->datatype)) == 0)
        {
            if ((uuid_compare(newcap.nugapp, target->nugapp)) == 0)
                return R_FAIL;

            /* Application type is different */
            uuid_copy(target->nugapp, newcap.nugapp);
        }
        else
        {
            if ((uuid_compare(newcap.nugapp, target->nugapp)) != 0)
                uuid_copy(target->nugapp, newcap.nugapp);
            uuid_copy(target->datatype, newcap.datatype);
        }

        if ((uuid_compare(target->nugtype, DETECTION)) == 0)
            addDefenseRouteEntry(target);

        addDataTrack(target);
        printRoutingTable();
        //printTypeTable();
        //printDataTable();
        return R_SUCCESS;
    }

    return R_FAIL;
}

/*
 *
 * EVENT ID TRANSLATION LOOKUP FUNCTIONS
 *
 */

//StoreEventLookup
//
//Takes a temporary eventid and does one of two things:
//
//   1) If an entry for that temp id does not currently exist
//      in the event lookup table, it will create one and
//      initialize it.
//
//   2) If an entry for that temp id ==DOES== exist, the function
//      will increment the current number of outstanding detection
//      requests matching that id
//
static HRESULT StoreEventLookup(unsigned eventid)
{

    HRESULT result;
    EVENTLOOKUPTABLE *lookup, **temp;

    result = FindEventLookup(eventid, &lookup);

    if (result == R_FOUND)
    {
        if (lookup->outstanding == ULONG_MAX)
            return R_FAIL;
        else
            lookup->outstanding++;
        return R_SUCCESS;
    }

    if ((lookup = calloc(1, sizeof(*lookup))) == NULL)
        return R_FAIL;
    lookup->tempid = eventid;
    lookup->outstanding = 1;

    for (temp = &lookuptable; *temp; temp=&(*temp)->next);
    *temp = lookup;

    return R_SUCCESS;
}

//FindEventLookup
//
//Finds an event lookup entry based on its temp id and returns a reference to it.
//
//This function is used only by the other event lookup functions
//to support their functionality.
//
static HRESULT FindEventLookup(unsigned tempid, EVENTLOOKUPTABLE **lookup)
{
    EVENTLOOKUPTABLE *temp;

    for (temp = lookuptable; temp; temp = temp->next)
    {
        if (tempid == temp->tempid)
        {
            *lookup = temp;
            return R_FOUND;
        }
    }

    *lookup = NULL;
    return R_NOT_FOUND;
}


//ChangeEventLookup
//
//Takes a temp id
//
//      If an entry exists in the lookup table matching that
//      temp id and it will decrement the number of outstanding
//      detection requests.
//
//      Should the number of outstanding requests reach 0,
//      that lookup entry is removed from the lookup
//      table entirely.
//
static HRESULT ChangeEventLookup(unsigned tempid)
{
    EVENTLOOKUPTABLE *lookup, *temp;
    HRESULT result;

    result = FindEventLookup(tempid, &lookup);

    if (result != R_FOUND)
        return result;

    lookup->outstanding--;

    if (lookup->outstanding == 0)
    {
        if (lookup == lookuptable)
            lookuptable = lookuptable->next;
        else
        {
            temp = lookuptable;
            while (temp->next != lookup)
                temp = temp->next;
            temp->next = lookup->next;
        }
    }
    else
        return R_SUCCESS;

    free(lookup);

    return R_SUCCESS;
}


//ResolveEventLookup
//
//Takes a temp id and a real id. Will behave different depending
//on the value of realid.
//
//   If realid is 0, the function will assume the realid is not
//   known and see if an entry matching that temp id exists
//   in the lookup table. If it does, the realid will point
//   to this value, otherwise a result code of R_NOT_FOUND
//   is returned.
//
//   If realid is non-zero, the function will set the real id
//   of the entry matching that temp id to the value pointed
//   to by realid. R_NOT_FOUND if no such entry exists
//
HRESULT ResolveEventLookup(unsigned tempid, unsigned *realid)
{
    EVENTLOOKUPTABLE *lookup;
    HRESULT result;

    result = FindEventLookup(tempid, &lookup);

    if (!*realid)
    {
        if (result == R_FOUND && lookup)
            *realid = lookup->realid;
    }
    else
    {
        if (result == R_FOUND && lookup)
        {
            lookup->realid = *realid;
            return R_SUCCESS;
        }
    }

    return result;
}



static HRESULT forwardDetectionData(RZBNetworkConnection *src, RZBNetworkConnection *dest, SNAKECHARMER *scpacket,
                                    REQPACKET **request, unsigned char **data, unsigned *size)
{
    REQPACKET *reqpacket;
    BLOCK_META_DATA *metaData;
    unsigned char *file = *data;

    /* If this is an initial request, populate reqpacket and event lookup table */
    if (*request == NULL)
    {
        reqpacket = malloc(sizeof(REQPACKET) + sizeof(BLOCK_META_DATA));
        *request = reqpacket;
        metaData = (BLOCK_META_DATA *)(reqpacket + 1);

        if (recvWrap(src, reqpacket, sizeof(*reqpacket)) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to receive request packet", "forwardDetectionData");
            return R_FAIL;
        }

        if (recvWrap(src, metaData, sizeof(*metaData)) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to receive file", "forwardDetectionData");
            return R_FAIL;
        }

        pthread_mutex_lock(&eventlookupmutex);
        StoreEventLookup(reqpacket->eventid);
        pthread_mutex_unlock(&eventlookupmutex);
    }
    else
    {
        reqpacket = *request;
        metaData = (BLOCK_META_DATA *)(reqpacket + 1);
    }

    // Connect to the destination
    // Initially this is set as REQ, but later on it is a different value
    scpacket->msgtype = REQ;

    if (sendWrap(dest, scpacket, sizeof(*scpacket)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send first snakecharmer packet", "forwardDetectionData");
        return R_FAIL;
    }

    if (sendWrap(dest, reqpacket, sizeof(*reqpacket)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send request packet", "forwardDetectionData");
        return R_FAIL;
    }

    if (sendWrap(dest, metaData, sizeof(*metaData)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send request packet", "forwardDetectionData");
        return R_FAIL;
    }

    if (recvWrap(dest, scpacket, sizeof(*scpacket)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to receive snakecharmer packet", "forwardDetectionData");
        return R_FAIL;
    }

    if (scpacket->msgtype != ACK)
    {
        submitLog(LOG_ERR, "Return message from detection nugget not ACK", "forwardDetectionData");
        return R_FAIL;
    }

    // If data != NULL, we've already got the data in memory
    if (!*data)
    {
        if (sendWrap(src, scpacket, sizeof(*scpacket)) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to send file transmit snakecharmer packet", "forwardDetectionDtaa");
            return R_FAIL;
        }

        if ((file = malloc(reqpacket->filesize)) == NULL)
        {
            submitLog(LOG_ERR, "Failed to allocate space for file pointer", "forwardDetectionData");
            return R_MALLOC_FAIL;
        }

        if (recvWrap(src, file, reqpacket->filesize) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to receive file", "forwardDetectionData");
            free(file);
            return R_FAIL;
        }

        scpacket->msgtype = RST;
        if (sendWrap(src, scpacket, sizeof(*scpacket)) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to send reset packet", "forwardDetectionData");
            return R_FAIL;
        }
        *data = file;
        *size = reqpacket->filesize;
    }

    if (sendWrap(dest, *data, *size) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send file", "forwardDeteectionData");
        return R_FAIL;
    }

    return R_SUCCESS;
}

/*
 *
 * Main Threaded function to handle Razorback request packets
 *
 */
static void *dispatchMaster(THREADARGS *threadargs)
{
    /* Routing variables */
    DATAENTRY *dataentry;
    APPENTRY *appentry;
    ROUTEENTRY *routeentry;
    NUGGETENTRY *nuggetcursor;
    NUGGETENTRY nuggetentry;
    SNAKECHARMER scpacket;
    REQPACKET *reqpacket;

    /* Tracking structures */
    unsigned char *data;
    unsigned size;
    char ip_str[INET6_ADDRSTRLEN];

    RZBNetworkConnection *net = threadargs->net;
    RZBNetworkConnection *net2;
    unsigned nuggetid = threadargs->nuggetid;

    unsigned tempid;
    HRESULT result;

    char port[33];

    //Receive first packet
    if (recvWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
    {
        perror("Did not receive first packet");
        NetworkConnectionDestroy(net);
        unthreadme(threadargs);
        return NULL;
    }

    //Ensure it's a razorback packet
    if (protocolCheck(&scpacket) == R_SUCCESS)
    {
        //Check the message type and call the associated handler
        switch (scpacket.msgtype)
        {
            case REG:
                //Register the nugget and hand back a nugget id
                if (handleRegistration(net, &nuggetid) != R_SUCCESS)
                    break;

                scpacket.msgtype = ACK;
                if (sendWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
                {
                    perror("Could not acknowledge nugget registration");
                    break;
                }

                if (sendWrap(net, &nuggetid, sizeof(nuggetid)) != R_SUCCESS)
                    perror("Could not send uuid");
                break;

            case NEWCAP:
                //Register the nugget and hand back a nugget id
                if ((handleNewCapability(net)) != R_SUCCESS)
                    break;

                scpacket.msgtype = ACK;
                if (sendWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
                {
                    perror("Could not acknowledge nugget registration");
                    break;
                }
                break;
            case REQBLOB:
                if ((handleAlertBlob(net)) != R_SUCCESS)
                    submitLog(LOG_ERR, "Failed to handle Alert Blob Request", "dispatchMater");
                break;

            case NTEL:
                if ((handleIntel(net)) != R_SUCCESS)
                    submitLog(LOG_ERR, "Failed in handleIntel", "dispatchMaster");
                break;

            case MD5CHECK:
                //Check the md5 against the dispatcher's local cache, then database
#ifdef RZB_PERF
                pthread_mutex_lock(&perfmutex);
                perf.db_cache_checks++;
                pthread_mutex_unlock(&perfmutex);
#endif /* RZB_PERF */
                if ((handleMD5CacheCheck(net)) != R_SUCCESS)
                    perror("Couldn not handle MD5 Cache Check\n");
                break;

            case ALRT:
                if ((handleAlert(net)) != R_SUCCESS)
                    perror("Failed to handle alerting");
                break;

            case META:
                if ((handleMeta(net)) != R_SUCCESS)
                    submitLog(LOG_ERR, "Failed to handle metadata", "dispatchmaster");
                break;

            case WKST:
                if ((handleWorkstation(net)) != R_SUCCESS)
                    perror("Failed to handle workstation");
                break;

            case VERDICT:
                if ((handleJudgement(net)) != R_SUCCESS)
                    submitLog(LOG_ERR, "Failed to handle verdict", "dispatchMaster");
                break;

            case REQ:
#ifdef RZB_PERF
                pthread_mutex_lock(&perfmutex);
                perf.routing_requests++;
                pthread_mutex_unlock(&perfmutex);
#endif /* RZB_PERF */
                //Receive and forward a detection request to a capable detection nugget
                uuid_copy(nuggetentry.datatype, scpacket.datatype);
                findDataEntry(&dataentry, &nuggetentry);

                // Data entry will be null if no nuggets are registered
                if (dataentry == NULL)
                {
#ifdef RZB_PERF
                    pthread_mutex_lock(&perfmutex);
                    perf.no_routing_table++;
                    pthread_mutex_unlock(&perfmutex);
#endif /* RZB_PERF */
                    sendFailure(net, &scpacket);
                    break;
                }

                // Prepare routing:
                // appentry loop is for each kind of nugget
                // targetting the data type.
                //
                // reqpacket is the local copy of the request packet, containing
                // the metadata for the block
                //
                // Data is the block of data passed, and size is the size of that data.
                //
                // We NULL / 0 these out as we use them to track if we've already received them.

                appentry = dataentry->apphead;
                if (!appentry)
                {
                    sendFailure(net, &scpacket);
                    break;
                }
                reqpacket = NULL;
                data = NULL;
                size = 0;

                while (appentry != NULL)
                {
                    routeentry = appentry->ondeck;
                    nuggetcursor = routeentry->nugget;
                    snprintf(port, sizeof(port), "%u", (unsigned)nuggetcursor->nugport);

                    // Update ondeck for round robin routing
                    if (routeentry->next == NULL)
                        appentry->ondeck = appentry->routehead;
                    else
                        appentry->ondeck = routeentry->next;

                    // Forward to the correct detection nugget
                    if (inet_ntop(AF_INET, &nuggetcursor->nugaddr, ip_str, sizeof(ip_str)) == NULL)
                        ip_str[0] = 0;
#ifdef RZB_PERF
                    pthread_mutex_lock(&perfmutex);
                    perf.data_forwards++;
                    pthread_mutex_unlock(&perfmutex);
#endif /* RZB_PERF */
                    if ((net2 = ConnectMe(ip_str, port)))
                    {
                        forwardDetectionData(net, net2, &scpacket, &reqpacket, &data, &size);
                        NetworkConnectionDestroy(net2);
                    }
                    else
                        printf("Failed to forward detection data to %s\n", ip_str);
                    appentry = appentry->next;
                }

                if (data != NULL) // Shouldn't be null, but just in case...
                    free(data);
                if (reqpacket != NULL) // Same deal
                    free(reqpacket);
                break;

            case DETECTDONE:
                if (recvWrap(net, &tempid, sizeof(tempid)) == R_FAIL)
                    perror("Failed to received tempid");

                result = ChangeEventLookup(tempid);

                if (result == R_NOT_FOUND)
                    printf("\nFailed to find tempid in lookup table\n");
                break;

            default:
                submitLog(LOG_ERR, "Invalid message type received", "dispatchMaster");
                printf("\nInvalid Message Type\n");
                scpacket.msgtype = RST;
                sendWrap(net, &scpacket, sizeof(scpacket));
                break;
        }
    }
    else
        printf("Failed the protocol check\n");

    pthread_mutex_lock(&perfmutex);
    unsigned long bytes_in;
    unsigned long bytes_out;
    NetworkConnectionGetStats(net, &bytes_in, &bytes_out);
    perf.bytes_in += bytes_in;
    perf.bytes_out += bytes_out;
    pthread_mutex_unlock(&perfmutex);
    NetworkConnectionDestroy(net);
    unthreadme(threadargs);

    return NULL;
}


void dispatchServer(const char *port)
{
    HRESULT retCode;

    /* Database */
    if (mysql_library_init(0, NULL, NULL))
    {
        perror("FATAL: mysql_library_init failure\n");
        exit (0);
    }

#ifdef RZB_PERF
    initPerf();
#endif /* RZB_PERF */

    /* Populate authorized nugget type table */
    retCode = buildUUIDList(&validNuggetCount, &validNuggetSet, "select UID from unique_ids where Metabook = 1");
    if (retCode != R_SUCCESS)
    {
        printf("Failed to build list of valid nuggets\n");
        exit(1);
    }
    printf("\n%u nugget types authorized:\n", validNuggetCount);
    printUUIDList(validNuggetSet, validNuggetCount);

    /* Populate handled data type table */
    retCode = buildUUIDList(&dataTypesCount, &dataTypes, "select UID from unique_ids where Metabook = 2");
    if (retCode != R_SUCCESS)
        exit(1);
    printf("\n%u Data types authorized:\n", dataTypesCount);
    printUUIDList(dataTypes, dataTypesCount);
    /*  End DB block */

    rzbServer(port, &dispatchMaster, 30);
    /* Cleanup database environement */
    printf("Server exiting....\n");
    printf("Closing DB connections\n");
    destroyMYSQLConnections();
    mysql_library_end();
#ifdef RZB_PERF
    printPerf();
#endif /* RZB_PERF */
    printf("Exiting...\n");
}

static HRESULT handleMD5CacheCheck(RZBNetworkConnection *net)
{
    MD5CACHEREQUEST request;
    MD5CACHERESPONSE response;
    SNAKECHARMER scpacket;
    BASIC_COMPONENT component;
    int cache_result;

    unsigned result;

    /* STATIC EVENT ID */
    static unsigned eventID = 1;

    if (recvWrap(net, &request, sizeof(request)) != R_SUCCESS)
    {
        perror("Could not get MD5 Cache Request information");
        return R_FAIL;
    }

    result = checkMD5(request.md5, request.filesize, &cache_result);
    if (result == R_NOT_FOUND)
    {
        memcpy(component.md5, request.md5, sizeof(component.md5));
        component.size = request.filesize;
        uuid_copy(component.datatype, request.datatype);
        insertBaseComponent(&component);
    }

    memset(&response, 0, sizeof(response));

    protocolFill(&scpacket);
    scpacket.msgtype = MD5RESPONSE;

    if (result == R_SUCCESS)
    {
        if (cache_result == CACHE_HIT_KNOWN_GOOD || cache_result == CACHE_HIT_KNOWN_BAD)
        {
            printf("Known hit!  Sending scpacket\n");
            if (sendWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
                return R_FAIL;

            //
            // This is probably unnecessary. We can combine these two HRESULT codes.
            //
            printf("Sending the cache response as known_good/bad\n");
            if (cache_result == CACHE_HIT_KNOWN_GOOD)
                response.result = KNOWN_GOOD_RESPONSE;
            else
                response.result = KNOWN_BAD_RESPONSE;

            if (sendWrap(net, &response, sizeof(response)) != R_SUCCESS)
                return R_FAIL;

            return R_SUCCESS;
        }
    }

    /* Fail closed -- Treat all responses that aren't explicit KNOWN_GOOD as R_NOT_FOUND
     * this way an error in another function does not interfere with detection */

    response.result = UNKNOWN_MD5_RESPONSE;

    pthread_mutex_lock(&eventidmutex);
    response.detectionID = eventID;
    eventID++;
    pthread_mutex_unlock(&eventidmutex);

    if (sendWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
        return R_FAIL;

    if (sendWrap(net, &response, sizeof(response)) != R_SUCCESS)
        return R_FAIL;

    return R_SUCCESS;
}

