#include <arpa/inet.h>
#include <uuid/uuid.h>
#include <time.h>
#include <stdio.h>

#include "config.h"
#include "pdf_dissector.h"
#include "rzb_global.h"
#include "rzb_client.h"
#include "rzb_network.h"
#include "rzb_alert_api.h"
#include "rzb_alert_global.h"
#include "rzb_alert_util.h"
#include "rzb_api.h"

#define MAX_ALERT_SIZE 512

static const char *DISSECTOR_PATH = "/usr/src/nrt/detection/pdf-dissector/dissector.py";

static void pdf_handler(BLOCK_META_DATA *metaData)
{
    char tmpFilePath[L_tmpnam], cmd[512];
    FILE *tmpFile = NULL, *pf = NULL;
    size_t len = MAX_ALERT_SIZE;
    ALERT alert;
    BLOCK_META_DATA *newMetaData = NULL;
    char *line = NULL;
    unsigned char tmp_md5[MD5_SIZE];
    char *msg = NULL;
    char *data_block = NULL;
    char *norm_block = NULL;
    char *short_data = NULL;
    char *long_data = NULL;

    printf("pdf_nugget received %u bytes of data\n", metaData->size);

    // Get our tmp file to work with
    if (tmpnam(tmpFilePath) == NULL)
    {
        printf("Error generating a temporary file name: %s\n", strerror(errno));
        return;
    }

    // Open the tmp file and write all of the data to it
    if ((tmpFile = fopen(tmpFilePath, "wb+")) == NULL)
    {
        printf("Error opening tmp file: %s\n", strerror(errno));
        return;
    }

    // Write our file and process the data
    if (fwrite(metaData->data, metaData->size, 1, tmpFile) != 1)
    {
        printf("Error while writing to tmp file %s: %s\n", tmpFilePath, strerror(errno));
        fclose(tmpFile);
    }
    else
    {

        // No need to keep this open
        fclose(tmpFile);

        // Build our cmd to run
        snprintf(cmd, sizeof(cmd) - 1, "%s %s", DISSECTOR_PATH, tmpFilePath);

        // Create our dissector
        if ((pf = popen(cmd, "r")) == NULL)
        {
            printf("Error while running dissector script\n");
        }
        else
        {

            // ALERT
            // <msg data>
            // UUID
            // <uuid of data>
            // DATA
            // <size of data>
            // <data>
            // NORM
            // <size of data>
            // <data>
            // SHORT
            // <size of data>
            // <data>
            // LONG
            // <size of data>
            // <data>
            // ENDALERT

            memset(&alert, 0, sizeof(alert));
            // Read all alerts generated
            while (!feof(pf))
            {
                if (getline(&line, &len, pf) > 0)
                {
                    if (!strncmp(line, "ALERT", 5))
                    {
                        memset(&alert, 0, sizeof(alert));

                        // Set up our static fields
                        alert.priority = 1;
                        if (inet_ntop(AF_INET, &metaData->src_ip, alert.src_ip, sizeof(alert.src_ip)) == NULL)
                            alert.src_ip[0] = 0;
                        if (inet_ntop(AF_INET, &metaData->dst_ip, alert.dst_ip, sizeof(alert.dst_ip)) == NULL)
                            alert.dst_ip[0] = 0;
                        alert.ip_proto = metaData->ip_proto;
                        alert.src_port = metaData->src_port;
                        alert.dst_port = metaData->dst_port;
                        alert.event_id = metaData->eventid;

                        // Read the next line which should have our alert text
                        if (getline(&line, &len, pf) > 0)
                        {
                            if (msg)
                                free(msg);
                            // Allocate storage for our message
                            msg = strdup(line);
                            alert.msg = msg;

                            // Set our msg size
                            alert.msg_size = strlen(msg);

                            // we dont' want the newline
                            if (msg[alert.msg_size - 1] == '\n')
                            {
                                msg[alert.msg_size - 1] = 0;
                                alert.msg_size--;
                            }
                        }
                    }
                    else if (!strncmp(line, "DATA", 4))
                    {
                        // Our next line should just be an integer value with our data length
                        if (fscanf(pf, "%u\n", &alert.db_size) != EOF)
                        {

                            // Make sure we actually have data to work with
                            if (alert.db_size > 0)
                            {
                                if (data_block)
                                    free(data_block);
                                // Allocate our storage for our data
                                if ((data_block = malloc(alert.db_size)) != NULL)
                                {
                                    // Read our data
                                    if (fread(data_block, alert.db_size, 1, pf) == 1)
                                    {
                                        // Add our md5 sum
                                        md5sum(data_block, alert.db_size, tmp_md5);
                                        alert.main_md5 = tmp_md5;
                                    }
                                    else
                                    {
                                        free(data_block);
                                        data_block = NULL;
                                    }
                                }
                                alert.data_block = data_block;
                            }
                        }
                    }
                    else if (!strncmp(line, "NORM", 4))
                    {
                        // Our next line should just be an integer value with our data length
                        if (fscanf(pf, "%u\n", &alert.norm_size) != EOF)
                        {
                            // Make sure we actually have data to work with
                            if (alert.norm_size > 0)
                            {
                                if (norm_block)
                                    free(norm_block);
                                // Allocate our storage for our data
                                if ((norm_block = malloc(alert.norm_size)) != NULL)
                                {
                                    // Read our data
                                    if (fread(norm_block, alert.norm_size, 1, pf) != 1)
                                    {
                                        free(norm_block);
                                        norm_block = NULL;
                                    }
                                }
                                alert.norm_block = norm_block;
                            }
                        }
                    }
                    else if (!strncmp(line, "SHORT", 5))
                    {
                        // Our next line should just be an integer value with our data length
                        if (fscanf(pf, "%u\n", &alert.sd_size) != EOF)
                        {
                            // Make sure we actually have data to work with
                            if (alert.sd_size > 0)
                            {
                                if (short_data)
                                    free(short_data);
                                // Allocate our storage for our data
                                if ((short_data = malloc(alert.sd_size)) != NULL)
                                {
                                    // Read our data
                                    if (fread(short_data, alert.sd_size, 1, pf) != 1)
                                    {
                                        free(short_data);
                                        short_data = NULL;
                                    }
                                }
                                alert.short_data = short_data;
                            }
                        }
                    }
                    else if (!strncmp(line, "LONG", 4))
                    {
                        // Our next line should just be an integer value with our data length
                        if (fscanf(pf, "%u\n", &alert.ld_size) != EOF)
                        {
                            // Make sure we actually have data to work with
                            if (alert.ld_size > 0)
                            {
                                if (long_data)
                                    free(long_data);
                                // Allocate our storage for our data
                                if ((long_data = malloc(alert.ld_size)) != NULL)
                                {
                                    // Read our data
                                    if (fread(long_data, alert.ld_size, 1, pf) != 1)
                                    {
                                        free(long_data);
                                        long_data = NULL;
                                    }
                                }
                                alert.long_data = long_data;
                            }
                        }
                    }
                    else if (!strncmp(line, "UUID", 4))
                    {
                        // Read the next line which should have our uuid
                        if (fread(line, SIZE_OF_UUID_STRING - 1, 1, pf) != 1)
                        {
                            perror("Error reading uuid");
                        }
                        else
                        {
                            // Add the null
                            line[SIZE_OF_UUID_STRING - 1] = 0x0;

                            // Assign our uuid
                            if (uuid_parse(line, alert.dataType) != 0)
                            {
                                perror("Error parsing uuid");
                            }
                        }

                        // We need to read the extra newline
                        if (fread(line, 1, 1, pf) != 1)
                            perror("Error reading uuid new line");
                    }
                    else if (!strncmp(line, "ENDALERT", 8))
                    {
                        // Send the alert and clean everything up
                        sendAlert(&alert);

                        // We may need to send component data back for processing
                        if (!uuid_compare(alert.dataType, PE_FILE))
                        {
                            // Copy our metadata for resubmission
                            if ((newMetaData = malloc(sizeof(*newMetaData))) == NULL)
                            {
                                perror("Error allocating space for new meta data block");
                            }
                            else
                            {
                                memcpy(newMetaData, metaData, sizeof(*newMetaData));

                                // Update the fieds that are required for us to send back
                                newMetaData->timestamp = time(NULL);
                                uuid_copy(newMetaData->datatype, PE_FILE);

                                // Try to use the normalized block if we have it
                                if (norm_block != NULL)
                                {
                                    newMetaData->data = (uint8_t *)norm_block;
                                    newMetaData->size = alert.norm_size;
                                    norm_block = NULL;
                                }
                                else if (data_block != NULL)
                                {
                                    newMetaData->data = (uint8_t *)data_block;
                                    newMetaData->size = alert.db_size;
                                    data_block = NULL;
                                }

                                // Make sure we aren't sending back the same data
                                if ((newMetaData->data != metaData->data) && (newMetaData->data != NULL))
                                {

                                    // Need to send the parent block as well
                                    newMetaData->parent_data = metaData->data;
                                    newMetaData->parent_size = metaData->size;
                                    memcpy(newMetaData->parent_md5, metaData->parent_md5, sizeof(newMetaData->parent_md5));

                                    // Finally send the data
                                    sendData(newMetaData);
                                }
                                else
                                {
                                    if (newMetaData->data)
                                        free(newMetaData->data);
                                    free(newMetaData);
                                }
                            }
                        }

                        if (msg)
                        {
                            free(msg);
                            msg = NULL;
                        }

                        if (data_block)
                        {
                            free(data_block);
                            data_block = NULL;
                        }

                        if (norm_block)
                        {
                            free(norm_block);
                            norm_block = NULL;
                        }

                        if (short_data)
                        {
                            free(short_data);
                            short_data = NULL;
                        }

                        if (long_data)
                        {
                            free(long_data);
                            long_data = NULL;
                        }
                    }
                }
            }
            pclose(pf);
        }
    }

    if (msg)
    {
        free(msg);
        msg = NULL;
    }

    if (data_block)
    {
        free(data_block);
        data_block = NULL;
    }

    if (norm_block)
    {
        free(norm_block);
        norm_block = NULL;
    }

    if (short_data)
    {
        free(short_data);
        short_data = NULL;
    }

    if (long_data)
    {
        free(long_data);
        long_data = NULL;
    }

    // Clean up
    if (remove(tmpFilePath) != 0)
    {
        printf("Error while deleting temp file %s", tmpFilePath);
    }
    if (line)
        free(line);
}

/*
 * Simple init function
 *
 * This function is run when your nugget lib is loaded and allows you to define what functions
 * should be called when certain datatypes are encountered by the system
 */
SO_PUBLIC HRESULT initNug(DetectionAPI *detectionObj)
{
    uuid_t list;

    uuid_copy(list, PDF_FILE);

    detectionObj->registerHandler(&pdf_handler, (const uuid_t *)&list, 1, TESTDETECT2);

    return(R_SUCCESS);
}

